#!/usr/bin/env python
'''
This file is part of HTD (HayzenTech Database)

Copyright (C) 2011 - 2012 Andrew Hayzen

HTD is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

HTD is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with HTD. If not, see <http://www.gnu.org/licenses/>

This document conforms to the PEP 8 Style Guide
'''

# Import built-in modules
from os import path as _path


#--------------------------------------------------------------------- Settings
__build_id__ = 310019
__version__ = 3.10


#------------------------------------------------------------------- Exceptions
class DatabaseError(Exception):
    pass


class AuthenticationError(DatabaseError):
    pass


class ConnectionError(DatabaseError):
    pass


class IntegrityError(DatabaseError):
    pass


class VersionError(DatabaseError):
    pass


#-------------------------------- Create containers for datatypes and filetypes
class _RegisterContainer():
    pass

DataType = _RegisterContainer()
FileType = _RegisterContainer()


#--------------------------------------------------------------- Register Class
class _Register(object):
    def __init__(self):
        self.dtype_codes = {}
        self.dtypes = {}
        self._default_dtype = None

        self.fhandlers = {}
        self._default_fhandler = None

    @property
    def default_dtype(self):
        return self._default_dtype

    @default_dtype.setter
    def default_dtype(self, dtype):
        if dtype.__name__ not in self.dtypes:
            raise ValueError("Default dtype must be a registered dtype.")

        self._default_dtype = dtype.__name__

    @property
    def default_fhandler(self):
        return self._default_fhandler

    @default_fhandler.setter
    def default_fhandler(self, fhandler):
        if fhandler not in self.fhandlers:
            raise ValueError("Default fhandler must be a registered fhandler.")

        self._default_fhandler = fhandler

    def register_dtype(self, dtype):
        setattr(DataType, dtype.__name__, dtype())

        self.dtype_codes[dtype.idn] = dtype.__name__
        self.dtypes[dtype.__name__] = dtype.idn

        return True

    def register_fhandler(self, fhandler, is_location, is_file=None):
        setattr(FileType, fhandler.__name__, fhandler)

        self.fhandlers[fhandler] = {"is_loc": is_location, "is_file": is_file}

        return True

    def resolve_fhandler(self, location):
        """ File Handler resolver - determins file type """

        # Filter fhandlers by location
        fhandlers = [f for f in self.fhandlers if
                     self.fhandlers[f]["is_loc"](location)]

        # Filter fhandlers by file
        output = []

        for f in fhandlers:
            # if file_format exists then uses File and Header standard
            if f.file_format is not None:
                if not _path.exists(location):
                    raise ConnectionError("No HTD file found (%s)" % location)

                # Open file and get fhandler code
                tmp = open(location, "rb")
                tmp.seek(8)
                if tmp.read(1) == f.file_format:
                    output.append(f)
                tmp.close()

            elif self.fhandlers[f]["is_file"] is not None:
                if self.fhandlers[f]["is_file"](location):
                    output.append(f)

        fhandlers = output

        # Load fhandler
        if len(fhandlers) == 0:
            raise ConnectionError("Unknown File Format")
        elif len(fhandlers) > 1:
            raise ConnectionError("Could not determine File Format either (%s)"
                                  % fhandlers)
        else:
            return fhandlers[0]

Register = _Register()

#---------------------------------------------------------------------- Imports

# Load objects
from ._base import Config, Progress
from ._base.trace import Event, EventType
from . import _base
from . import _workers
from ._objects.attribute import Attribute
from ._objects.user import Authenticate, User
from ._objects.connection import Connection as _Connection


#------------------------------------------------------ Top level HTD functions
@_base.trace.trace
def connect(location, user=None, file_obj=None):
    """ Connect to a table or db """
    return _Connection(location, user, file_obj)


@_base.trace.trace
def create(location, identifier=None, attributes={}, users={}, login_user=None,
           htd_file=None):
    """
    Create a new HTD File
    To create a table parse the identifier for the table otherwise leave
    the identifier blank to create a db
    """

    # Confirm location is str
    if not isinstance(location, str):
        raise ValueError("location (%s) must be a str object." % location)

    # Confirm attributes is dict
    if not isinstance(attributes, dict):
        raise ValueError("attributes (%s) must be a dict object." % attributes)

    # Confirm users is dict
    if not isinstance(users, dict):
        raise ValueError("users (%s) must be a dict object." % users)

    # Check for identifier
    if identifier is not None:
        if not isinstance(identifier, list):  # Convert single into a list
            identifier = [identifier]

        # Create ids
        for i in identifier:
            if i not in attributes:
                raise ValueError("Identifier must be in attributes parameter.")

        x = 0

        # Check Attributes are an Attribute object
        for i in attributes:
            if not isinstance(attributes[i], Attribute):
                raise ValueError("Attribute(s) must be an Attribute object.")

            attributes[i]._id = x  # Force IDs
            x += 1

            if i is None:
                raise ValueError("Attribute name cannot be NULL.")

            attributes[i].name = i  # Run name check

        table = None  # Set TABLE mode
    else:
        # Create db attributes
        name_att = Attribute(DataType.STRING)
        name_att._id = 0
        loc_att = Attribute(DataType.STRING)
        loc_att._id = 1

        # Load attribute and identifier
        attributes = {"name": name_att, "location": loc_att}
        identifier = ["name"]
        table = False  # Set DB mode master file

    # Load guest account if needed
    guest = False
    if len(users) == 0:
        users = {"guest": User(None)}
        guest = True

    # Confirm users are valid
    acc = False

    for i in users:
        if not isinstance(users[i], User):  # Error if not User object
            raise ValueError("User(s) must be an User object.")

        if users[i].access == "roadx":  # Check for full access
            acc = True

        users[i].username = i  # Add username check

    if not acc:  # At least one user has full access
        raise ValueError("One User must have full access (roadx)")

    if len(users) > 1:
        if login_user is None:
            raise ValueError("A login_user must be specified")
    else:
        login_user = list(users.keys())[0]

    if login_user not in users:
        raise ValueError("login_user must be in users")
    else:
        if guest:
            cred = Authenticate("guest", None)
        else:
            cred = Authenticate(user_obj=users[login_user])

    # Load file type
    if htd_file is None:
        htd_file = Register.default_fhandler

    # Check htd_file exists
    if htd_file not in Register.fhandlers:
        raise ValueError("htd_file has not been registered.")

    # Check location is correct for file type
    if Register.fhandlers[htd_file]["is_loc"] is not None:
        if not Register.fhandlers[htd_file]["is_loc"](location):
            raise ValueError("An invalid location (%s) was given for the " \
                             "file format (%s)" % (location, htd_file))

    # Build Header
    header = {"TAB": table, "ATT": attributes, "USR": users, "KEY": identifier,
              "VER": __version__, "R1M": {}, "RM1": {}}

    # Load and set header
    f = htd_file(location, header)

    if f.persistent:  # Close the file if persistent
        f.close()
        f = None

    return connect(location, cred, file_obj=f)  # Return con obj

Event.add(None, "create")  # Manually add create
Event.add(None, "connect")  # Manually add connect
