'''
This file is part of HTD (HayzenTech Database)

Copyright (C) 2011 - 2012 Andrew Hayzen

HTD is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

HTD is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with HTD. If not, see <http://www.gnu.org/licenses/>
'''


class _Config(object):
    """ Config for HTD with auto attach to Event """
    errors = True
    force_sync = False
    _debug = False
    _progress = False

    @property
    def debug(self):
        return self._debug

    @debug.setter
    def debug(self, state):
        if state:
            Evt.connect(EvtType.BEFORE_RUN, EvtType.ALL_SIGNALS, call)
            Evt.connect(EvtType.AFTER_RUN, EvtType.ALL_SIGNALS, returned)
            Evt.connect(EvtType.ERROR, EvtType.ALL_SIGNALS, error)
        else:
            Evt.disconnect(EvtType.BEFORE_RUN, EvtType.ALL_SIGNALS, call)
            Evt.disconnect(EvtType.AFTER_RUN, EvtType.ALL_SIGNALS, returned)
            Evt.disconnect(EvtType.ERROR, EvtType.ALL_SIGNALS, error)

        self._debug = state

    @property
    def progress(self):
        return self._progress

    @progress.setter
    def progress(self, state):
        arr = ((EvtType.BEFORE_LOCK, EvtType.ALL_SIGNALS, prog_create),
               (EvtType.BEFORE_RUN, EvtType.ALL_SIGNALS, prog_brun),
               (EvtType.AFTER_RUN, EvtType.ALL_SIGNALS, prog_arun),
               (EvtType.AFTER_UNLOCK, EvtType.ALL_SIGNALS, prog_remove),
               (EvtType.ERROR, EvtType.ALL_SIGNALS, prog_remove))

        if state:
            for args in arr:
                Evt.connect(*args)
        else:
            for args in arr:
                Evt.disconnect(*args)


Config = _Config()  # Trace requires Config

from .. import AuthenticationError
from .file import File
from .trace import Event as Evt, EventType as EvtType
from . import types

from copy import copy
from os import path
from threading import current_thread, Lock

perms = {"r": "read", "o": "overwrite", "a": "append",
         "d": "delete", "x": "admin"}


def convert_location(location, add):
    """ Insert *add* into the location (/a/b.htd => /a/.~lock_b.htd) """
    folder = location.split(path.sep)
    name = "".join([add, folder[-1]])  # create new filename

    folder.pop()  # remove name
    folder.append(name)

    return path.sep.join(folder)


def location_is_file(location):
    """ Check location is a HTD file """
    if not path.isabs(location):
        return False

    directory, name = path.split(location)

    if not path.isdir(directory):
        return False

    if not location.endswith(".htd"):
        return False

    return True


def perm(level, access=""):
    """ Run access permission check """

    perm = [i for i in level if i not in access]

    if len(perm) == 0:
        return True

    err = ", ".join(['(%s - %s)' % (i, perms[i]) for i in perm])
    err = 'User doesn\'t have to permission [%s]' % err

    raise AuthenticationError(err)


class ThreadLock(object):
    def __init__(self):
        self._threading_lock = Lock()

    def _acquire(self):
        """ Acquire threading lock """
        return self._threading_lock.acquire()

    def _release(self):
        """ Release threading lock """
        return self._threading_lock.release()

    @property
    def _thread_locked(self):
        return self._threading_lock.locked()


#--------------------------------------------------------- Callbacks for config
def call(idn, event, signal, *args, **kwargs):
    # Build parameters
    tmp = [str(i) for i in args]
    tmp.extend(["%s=%s" % (i, str(kwargs[i])) for i in kwargs])
    param = ", ".join(tmp)

    print("HTD DEBUG: Calling %s(%s)" % (signal, param))


def returned(idn, event, signal, result, *args, **kwargs):
    print("HTD DEBUG: %s returned %s" % (signal, result))


def error(idn, event, signal, e, *args, **kwargs):
    txt = "%s, %s (%s)" % (type(e).__name__, e, signal)

    print("HTD DEBUG: %s" % txt)


def prog_create(idn, event, signal, *args, **kwargs):
    Progress.add(idn, ProgressItem(current_thread(), 3,
                                   "Waiting for lock (%s)" % signal))


def prog_brun(idn, event, signal, *args, **kwargs):
    p = Progress.get(idn)
    p.count()
    p.text = "Calling %s" % signal


def prog_arun(idn, event, signal, *args, **kwargs):
    p = Progress.get(idn)
    p.count()
    p.text = "%s Returned" % signal


def prog_remove(idn, event, signal, *args, **kwargs):
    Progress.remove(idn)


#-------------------------------------------------- Progress management for HTD
class ProgressItem():
    """ A progression Item """
    def __init__(self, thread, length, text):
        self.thread = thread
        self.length = length
        self.text = text
        self.counter = 0

    def count(self, n=1):
        self.counter += n


class ProgressItems():
    """ Progress management """
    def __init__(self):
        self.tasks = {}
        self.threads = {}

    def add(self, idn, progress):
        """ Add new progress item """
        if not isinstance(progress, ProgressItem):
            raise ValueError("Must be a ProgressItem instance")

        self.tasks[idn] = progress

        if progress.thread not in self.threads:
            self.threads[progress.thread] = []

        self.threads[progress.thread].append(progress)

    def get(self, idn):
        return self.tasks[idn]

    def progress(self):
        """ Return all running tasks """

        return [job for job in copy(self.tasks.values())]

    def remove(self, idn):
        try:
            p = self.tasks[idn]
            self.threads[p.thread].remove(p)
        except KeyError:
            pass

        try:
            del self.tasks[idn]
        except KeyError:
            pass

    def thread_progress(self, thread):
        """ Return the running tasks for a specific thread """
        if thread not in self.threads:
            raise KeyError("Thread does not exist")

        return [{"length":job.length, "count":job.counter, "name":job.text,
                "fraction":job.counter / job.length} for job in
               copy(self.threads[thread])]

Progress = ProgressItems()
