'''
This file is part of HTD (HayzenTech Database)

Copyright (C) 2011 - 2012 Andrew Hayzen

HTD is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

HTD is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with HTD. If not, see <http://www.gnu.org/licenses/>

Base File for HTD
Provides slicing and file locking support for files
'''

import os

if os.name == "posix":
    import fcntl
elif os.name == "nt":
    import win32con
    import win32file
    import pywintypes
else:
    print("HTD Warning: Unknown operating system, FileLock unavailable.")


class File(object):
    """ HTD File that provides slicing and file locking """
    def __init__(self, location):
        self.__file = open(location, "rb+")
        self.__lkcur = False

        # Add file methods to self
        for i in [i for i in dir(self.__file) if '_' not in i and
                  i != "close"]:
            setattr(self, i, getattr(self.__file, i))

    def __delitem__(self, key):
        # Delete Slice (MUST be a slice)
        self.seek(key.stop)
        text = self.read()
        self.seek(key.start)
        self.write(text)
        self.truncate()
        self.flush()

    def __getitem__(self, key):
        # Return slice (MUST be a slice)
        self.seek(key.start, 0)
        return self.read((key.stop - key.start))

    def __len__(self):
        self.seek(0, 2)
        return self.tell()

    def __setitem__(self, key, value):
        # Slice (MUST be a slice)
        i = key.start
        e = key.stop
        diff = e - i
        if (diff != len(value)):
            self.seek(e)
            text = self.read()
            self.seek(i)
            self.write(value)
            self.write(text)
            self.truncate()
        else:
            self.seek(i)
            self.write(value)
        self.flush()

    if os.name == "posix":
        def _lock_reading(self):
            """ Lock file for reading """

            fcntl.flock(self.__file, fcntl.LOCK_SH | fcntl.LOCK_NB)
            self.__lkcur = None

            return True

    elif os.name == "nt":
        def _lock_reading(self):
            """ Lock file for reading """

            self._unlock()  # NT requires unlock first
            f = win32file._get_osfhandle(self.fileno())
            try:
                win32file.LockFileEx(f, 0 | win32con.LOCKFILE_FAIL_IMMEDIATELY,
                                      0, -0x10000, pywintypes.OVERLAPPED())
            except pywintypes.error as exc_value:
                if exc_value[0] == 33:  # File locked exception
                    raise IOError("File is locked")
                else:
                    raise IOError("Unknown Error")

            self.__lkcur = None

            return True

    if os.name == "posix":
        def _lock_writing(self):
            """ Lock file for writing """

            fcntl.flock(self.__file, fcntl.LOCK_EX | fcntl.LOCK_NB)
            self.__lkcur = True

            return True

    elif os.name == "nt":
        def _lock_writing(self):
            """ Lock file for writing """

            self._unlock()  # NT requires unlock first
            f = win32file._get_osfhandle(self.fileno())
            try:
                win32file.LockFileEx(f, win32con.LOCKFILE_EXCLUSIVE_LOCK |
                                     win32con.LOCKFILE_FAIL_IMMEDIATELY, 0,
                                     - 0x10000, pywintypes.OVERLAPPED())
            except pywintypes.error as exc_value:
                if exc_value[0] == 33:  # File locked exception
                    raise IOError("File is locked")
                else:
                    raise IOError("Unknown Error")

            self.__lkcur = True

            return True

    if os.name == "posix":
        def _unlock(self):
            """ Unlock file """

            fcntl.flock(self.__file, fcntl.LOCK_UN)
            self.__lkcur = False

            return True

    elif os.name == "nt":
        def _unlock(self):
            """ Unlock file """

            f = win32file._get_osfhandle(self.fileno())
            try:
                win32file.UnlockFileEx(f, 0, -0x10000, pywintypes.OVERLAPPED())
            except pywintypes.error as exc_value:
                if exc_value[0] == 158:  # Don't raise already locked
                    pass
                else:
                    raise

            self.__lkcur = False

            return True

    def close(self):
        self.flush()
        try:
            self._unlock()
        except IOError:
            pass
        self.__file.close()

        return True

    @property
    def lock_type(self):
        return self.__lkcur
