'''
This file is part of HTD (HayzenTech Database)

Copyright (C) 2011 - 2012 Andrew Hayzen

HTD is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

HTD is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with HTD. If not, see <http://www.gnu.org/licenses/>

Tracing for HTD
Captures progress, logging, triggers, errors and permission levels
'''

from .._base import Config
from inspect import getsourcelines
from threading import Lock


#----------------------------------------------------- Event Management for HTD
class EventType():
    """ Event types """
    BEFORE_LOCK = 1
    BEFORE_RUN = 2
    ERROR = 3
    AFTER_RUN = 4
    AFTER_UNLOCK = 5
    ALL_EVENTS = 6
    ALL_SIGNALS = 7


class Monitor():
    """ Class that emits and connects event """
    def __init__(self):
        self.signals = {7: {}}  # Signals and events

        # Load events for ALL_SIGNALS
        for i in range(1, 7):
            self.signals[7][i] = []

    def add(self, cls, func):
        if cls is None:
            signal = func
        else:
            signal = ".".join([cls, func])  # Make signal

        # Load signal
        if signal not in self.signals:
            self.signals[signal] = {}

        # Load events for ALL_SIGNALS
        for i in range(1, 7):
            if i not in self.signals[signal]:
                self.signals[signal][i] = []

    def connect(self, event, signal, func):

        try:
            if func not in self.signals[signal][event]:
                self.signals[signal][event].append(func)
        except KeyError:
            raise KeyError("Could not connect func as no event or signal " \
                           "existed (%s, %s)" % event, signal)

        return

    def disconnect(self, event, signal, func):

        try:
            if func in self.signals[signal][event]:
                self.signals[signal][event].remove(func)
        except KeyError:
            raise KeyError("Could not connect func as no event or signal " \
                           "existed (%s, %s)" % event, signal)

        return

    def emit(self, idn, event, signal, *args, **kwargs):
        for i in self.signals[7][event]:  # Run ALL_SIGNALS
            i(idn, event, signal, args, kwargs)

        for i in self.signals[7][6]:  # Run ALL_EVENTS ALL_SIGNALS
            i(idn, event, signal, args, kwargs)

        for i in self.signals[signal][6]:  # Run ALL_EVENTS
            i(idn, event, signal, args, kwargs)

        for i in self.signals[signal][event]:  # Run Normal
            i(idn, event, signal, args, kwargs)

Event = Monitor()


def list_dec_methods(cls, name):
    """ Lists methods in a class with a specific decorator """
    sourcelines = getsourcelines(cls)[0]

    for i, line in enumerate(sourcelines):
        line = line.strip()

        if line == "".join(["@", name]):
            next_line = sourcelines[i + 1]
            yield next_line.split('def')[1].split('(')[0].strip()


def register(func):
    """ Registers a class with HTD events """
    cls = func.__name__

    for i in list_dec_methods(func, "trace"):
        Event.add(cls, i)

    return func


# Event ids
idn_lock = Lock()
idn = 1


def trace(func):
    """ Trace function  - Manages threads and emits events """

    def callf(*args, **kwargs):
        global idn

        # Generate name
        func_name = func.__name__

        idn_lock.acquire()
        i = idn
        idn += 1
        idn_lock.release()

        # Check that object isn't closed
        try:
            if not args[0].alive:
                return None
        except (IndexError, AttributeError):
            tmp = None
        else:
            tmp = args[0]
            func_name = ".".join([tmp.__class__.__name__, func_name])

        Event.emit(i, 1, func_name, args, kwargs)

        # Acquire lock
        locked = False

        try:
            if not tmp._thread_locked:
                tmp._acquire()
                locked = True
        except AttributeError:
            pass

        Event.emit(i, 2, func_name, args, kwargs)

        try:
            result = func(*args, **kwargs)
        except Exception as e:
            # Unlock thread and file lock
            if (hasattr(tmp, "_file") and hasattr(tmp._file, "_t_locked") and
                tmp._file._t_locked):
                tmp._file.unlock()

            if hasattr(tmp, "_thread_locked") and tmp._thread_locked:
                tmp._release()

            Event.emit(i, 3, func_name, e, args, kwargs)

            if Config.errors:
                raise

            return False
        else:
            Event.emit(i, 4, func_name, result)

        if locked:  # Release lock this locked it
            try:
                if tmp._thread_locked:
                    tmp._release()
            except AttributeError:
                pass

        Event.emit(i, 5, func_name, result)

        return result

    return callf
