'''
This file is part of HTD (HayzenTech Database)

Copyright (C) 2011 - 2012 Andrew Hayzen

HTD is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

HTD is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with HTD. If not, see <http://www.gnu.org/licenses/>

HTD Data Types
Data types and (un)sanisite functions are managed here
'''

from datetime import date, datetime, time, timedelta
from .. import __version__, DataType, Register
from struct import error as struct_error, pack, unpack
import sys


#-------------------------------------------------------------------- HTD Types
class DType(object):
    def __repr__(self):
        return ("<HTD%s Data Type (%s)>" %
                (__version__, self.__class__.__name__))

reg_dtype = Register.register_dtype  # Register dtypes


#--------------------------------------------------------------- Raw data types
class BOOL(DType):
    idn = b"\x01"
    can_range = False

    def sanitise(self, real):
        if real:
            return b"1"
        else:
            return b"0"

    size = 1

    def type_check(self, real):
        return type(real) is bool

    def unsanitise(self, raw):
        if raw == b"1":
            return True
        else:
            return False

reg_dtype(BOOL)


class BYTES(DType):
    idn = b"\x02"
    can_range = False
    sanitise = bytes
    size = 0

    def type_check(self, real):
        return type(real) is bytes

    unsanitise = bytes

reg_dtype(BYTES)


class STRING(DType):
    idn = b"\x04"
    can_range = False

    if sys.version_info < (3, 0):
        sanitise = str
    else:
        def sanitise(self, raw):
            return raw.encode("ISO-8859-1")

    size = 0

    def type_check(self, real):
        return type(real) is str

    if sys.version_info < (3, 0):
        unsanitise = str
    else:
        def unsanitise(self, raw):
            return raw.decode("ISO-8859-1")

reg_dtype(STRING)


#----------------------------------------------- Date and time based data types
class DATE(DType):
    idn = b"\x07"
    can_range = False

    def sanitise(self, raw):
        return raw.isoformat().encode("ISO-8859-1")

    size = 9

    def type_check(self, real):
        return type(real) is date

    def unsanitise(self, raw):
        data = raw.decode("ISO-8859-1").split("-")
        return date(int(data[0]), int(data[1]), int(data[2]))

reg_dtype(DATE)


class TIME(DType):
    idn = b"\x08"
    can_range = False

    def sanitise(self, raw):
        return raw.isoformat().encode("ISO-8859-1")

    size = 8

    def type_check(self, real):
        return type(real) is time

    def unsanitise(self, raw):
        data = raw.decode("ISO-8859-1").split(":")
        data[2] = data[2].split(".")
        data[2].append(0)
        return time(int(data[0]), int(data[1]), int(data[2][0]),
                    int(data[2][1]))

reg_dtype(TIME)


class DATETIME(DType):
    idn = b"\t"
    can_range = False

    def sanitise(self, raw):
        return raw.isoformat().encode("ISO-8859-1")

    size = 26

    def type_check(self, real):
        return type(real) is datetime

    def unsanitise(self, raw):
        data = raw.decode("ISO-8859-1").split(".")
        dt = datetime.strptime(data[0], "%Y-%m-%dT%H:%M:%S")

        if len(data) > 1:
            us = int(data[1].rstrip("Z"), 10)
        else:
            us = 0

        return dt + timedelta(microseconds=us)

reg_dtype(DATETIME)


#------------------------------------------------------------ Number data types
class NumBase(DType):
    can_range = True

    def sanitise(self, real):
        return pack(self._c, real)

    def unsanitise(self, raw):
        return unpack(self._c, raw)[0]


class BYTE(NumBase):
    idn = b"\x0f"
    _c = "b"
    can_range = True
    size = 1

    def type_check(self, real):
        return type(real) is int


class UBYTE(NumBase):
    idn = b"\x10"
    _c = "B"
    can_range = True
    size = 1

    def type_check(self, real):
        return type(real) is int


class SHORT(NumBase):
    idn = b"\x11"
    _c = "h"
    can_range = True
    size = 2

    def type_check(self, real):
        return type(real) is int


class USHORT(NumBase):
    idn = b"\x12"
    _c = "H"
    can_range = True
    size = 2

    def type_check(self, real):
        return type(real) is int


class INTEGER(NumBase):
    idn = b"\x13"
    _c = "i"
    can_range = True
    size = 4

    def type_check(self, real):
        return type(real) is int


class UINTEGER(NumBase):
    idn = b"\x14"
    _c = "I"
    size = 4

    def type_check(self, real):
        return type(real) is int


class LONG(NumBase):
    idn = b"\x15"
    _c = "q"
    size = 8

    if sys.version_info < (3, 0):
        def type_check(self, real):
            return type(real) is long
    else:
        def type_check(self, real):
            return type(real) is int


class ULONG(NumBase):
    idn = b"\x16"
    _c = "Q"
    size = 8

    if sys.version_info < (3, 0):
        def type_check(self, real):
            return type(real) is long
    else:
        def type_check(self, real):
            return type(real) is int


class FLOAT(NumBase):
    idn = b"\x1e"
    _c = "f"
    size = 4

    def type_check(self, real):
        return type(real) is float


class DFLOAT(NumBase):
    idn = b"\x1f"
    _c = "d"
    size = 8

    def type_check(self, real):
        return type(real) is float


reg_dtype(BYTE)
reg_dtype(UBYTE)
reg_dtype(SHORT)
reg_dtype(USHORT)
reg_dtype(INTEGER)
reg_dtype(UINTEGER)
reg_dtype(LONG)
reg_dtype(ULONG)
reg_dtype(FLOAT)
reg_dtype(DFLOAT)


#------------------------------------------------- Python 3 specific data types
if sys.version_info < (3, 0):
    class UTF8(DType):
        idn = b"#"
        can_range = False

        def sanitise(self, real):
            return real.encode("utf8")

        size = 0

        def type_check(self, real):
            return type(real) is unicode

        def unsanitise(self, raw):
            return raw.decode("utf8")

    reg_dtype(UTF8)

    class UTF16(DType):
        idn = b"$"
        can_range = False

        def sanitise(self, real):
            return real.encode("utf16")

        size = 0

        def type_check(self, real):
            return type(real) is unicode

        def unsanitise(self, raw):
            return raw.decode("utf16")

    reg_dtype(UTF16)

    class UTF32(DType):
        idn = b"%"
        can_range = False

        def sanitise(self, real):
            return real.encode("utf32")

        size = 0

        def type_check(self, real):
            return type(real) is unicode

        def unsanitise(self, raw):
            return raw.decode("utf32")

    reg_dtype(UTF32)


#----------------------------------------------------------- Preserve data type
class PRESERVE(DType):
    idn = b"\x0b"
    can_range = False

    def sanitise(self, real):
        for ty in Register.dtype_codes:
            x = getattr(DataType, Register.dtype_codes[ty])
            if isinstance(x, PRESERVE) is False and x.type_check(real):
                try:
                    tmp = b"".join([x.sanitise(real), ty])
                except struct_error:
                    continue
                else:
                    return tmp

        raise ValueError("Unknown type to preserve (%s)" % type(real))

    size = 0

    def type_check(self, real):
        for ty in Register.dtype_codes:
            x = getattr(DataType, Register.dtype_codes[ty])
            if isinstance(x, PRESERVE) is False and x.type_check(real):
                return True

        return False

    if sys.version_info < (3, 0):
        def unsanitise(self, raw):
            ty = raw[-1]
            raw = raw[:-1]

            if ty in Register.dtype_codes:
                x = getattr(DataType, Register.dtype_codes[ty])
            else:
                raise TypeError("Could not restore type.")

            return x.unsanitise(raw)
    else:
        def unsanitise(self, raw):
            ty = chr(raw[-1]).encode("ISO-8859-1")
            raw = raw[:-1]

            if ty in Register.dtype_codes:
                x = getattr(DataType, Register.dtype_codes[ty])
            else:
                raise TypeError("Could not restore type.")

            return x.unsanitise(raw)


reg_dtype(PRESERVE)
Register.default_dtype = PRESERVE
