'''
This file is part of HTD (HayzenTech Database)

Copyright (C) 2011 - 2012 Andrew Hayzen

HTD is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

HTD is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with HTD. If not, see <http://www.gnu.org/licenses/>

HTD Attribute
'''

from .. import __version__, Register, DataType
from .._workers import sanitise, unsanitise
from struct import pack


class Attribute(object):
    """ Class that manages a specific attribute """
    def __init__(self, dtype=None, size=None):
        # Save variables
        self.dtype = dtype
        self.size = size

        # Set extra options
        self._default = None
        self._id = None
        self._limit_list = [False, []]
        self._name = None
        self._range = False
        self._re = None
        self._required = False

    def __repr__(self):
        return ("<HTD%s Attribute (<name '%s'> <type '%s'> <len %s>)>" %
                (__version__, self.name, self.dtype.__class__.__name__,
                 self.size))

    @property
    def default(self):
        return unsanitise({"val": self._default}, {"val": self})["val"]

    @default.setter
    def default(self, value):
        value = sanitise({"val": value}, {"val": self})
        if not value:
            self._default = None
        else:
            self._default = value["val"]

    @property
    def dtype(self):
        return self._dtype

    @dtype.setter
    def dtype(self, dtype):
        # Check type
        if dtype is None:
            dtype = Register.default_dtype
        else:
            dtype = dtype.__class__.__name__

        try:
            dtype = getattr(DataType, dtype)
        except AttributeError:
            raise ValueError("Attribute type must be a valid HTD data type")

        self._dtype = dtype

    @property
    def limit_list(self):
        return self._limit_list

    @limit_list.setter
    def limit_list(self, limits):
        if limits is False or limits == [False, []]:
            self._limit_list = [False, []]
        elif isinstance(limits, (list, tuple)) is False:
            raise ValueError("Limits must be list type")
        else:
            self._limit_list = [True, limits]

    @property
    def name(self):
        return self._name

    @name.setter
    def name(self, name):
        # Check name
        if name is None:
            self._name = None
            return

        if isinstance(name, str) is False:
            raise ValueError("Attribute name must be str type.")

        if "." in name:
            raise ValueError("Attribute name must not contain '.'")

        self._name = name

    @property
    def range(self):
        return self._range

    @range.setter
    def range(self, ran):
        if ran is False or ran == [False, False]:
            self._range = ran
            return

        if isinstance(ran, (list, tuple)) is False:
            raise ValueError("Range must be list type")

        lower, upper = ran

        if lower >= upper:
            raise ValueError("The lower value must be less than the upper.")

        self._range = [lower, upper]

    @property
    def reg_expr(self):
        return self._re

    @reg_expr.setter
    def reg_expr(self, expr):
        if type(expr) is not bytes and not not expr:
            expr = str(expr).encode("ISO-8859-1")
        self._re = expr

    @property
    def required(self):
        return self._required

    @required.setter
    def required(self, required):
        if required in [True, False]:
            self._required = required
        else:
            raise ValueError("Required must be either True or False")

    @property
    def size(self):
        return self._size

    @size.setter
    def size(self, size):
        # Check dtype restrictions
        if self._dtype is not None:
            if self._dtype.size != 0:
                size = self._dtype.size

        # Check size
        if size is None:
            self._size = 255
            return

        if isinstance(size, int) is False:
            raise ValueError("Attribute size must be an Integer")
        else:
            if not 1 <= size < 4294967295:
                raise ValueError("Attribute size must be greater than or " \
                                 "equal to 1 and less than 4294967295")

        self._size = size

    def to_dict(self):
        """ Convert to string """

        return {"dtype": self.dtype.idn, "id": self._id,
                "limit_list": self._limit_list, "default": self._default,
                "range": self._range, "reg_expr": self._re,
                "required": self._required, "size": pack("I", self._size)}
