'''
This file is part of HTD (HayzenTech Database)

Copyright (C) 2011 - 2012 Andrew Hayzen

HTD is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

HTD is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with HTD. If not, see <http://www.gnu.org/licenses/>

HTD Attributes
'''

from .. import ConnectionError, IntegrityError, Register, __version__
from .._base import perm, ThreadLock
from .._base.trace import trace, register
from .._objects.attribute import Attribute
from .._workers import sanitise, unsanitise
from weakref import ref


@register
class Attributes(ThreadLock):
    """ Class that manages attribute settings in a table """
    def __init__(self, parent, location, file_obj=None):
        # Set vars
        self.alive = True
        self._parent = ref(parent)
        self.connection._children.append(self)

        self._access = self.connection._access  # Load user access

        self._index = 0  # var for loop

        if file_obj is None:
            htdfile = Register.resolve_fhandler(location)
            self._file = htdfile(location)
            self._opened = True
        else:
            self._file = file_obj
            self._opened = False

        ThreadLock.__init__(self)

    @trace
    def __contains__(self, key):
        """ Return True if attribute exists in db """

        perm("r", self._access)

        self._file.lock_reading()

        out = False
        if key in self._file.header["ATT"]:
            out = True

        self._file.unlock()

        return out

    @trace
    def __delitem__(self, key):
        """ Delete attribute and data from db """

        perm("d", self._access)

        self._file.lock_writing()

        # Confirm att exists
        if key not in self._file.header["ATT"]:
            raise KeyError('Could not find attribute to delete. (%s)' % key)

        if key in self._file.header["KEY"]:
            raise ValueError('Cannot remove attribute that is an identifier')

        if key in self._file.header["R1M"] or key in self._file.header["RM1"]:
            raise IntegrityError("Relationship(s) must be broken first")

        self._file.delete_attribute(key)

        # Delete attribute, rebuild header
        header = self._file.header

        i = header["ATT"][key]._id

        del header["ATT"][key]

        # resort IDs
        for a in header["ATT"]:
            att = header["ATT"][a]
            if att._id > i:
                att._id -= 1

        self._file.header_update(header)

        self._file.unlock()

    @trace
    def __getitem__(self, key):
        """ Return attributes settings """

        perm("r")

        self._file.lock_reading()

        attributes = self._file.header["ATT"]

        if key not in attributes:
            raise KeyError('Failed to find sanitise settings for attribute ' \
                           '(%s)' % (key))

        self._file.unlock()

        return attributes[key]

    @trace
    def __iter__(self):
        perm("r", self._access)

        self._index = 0

        return self

    @trace
    def __len__(self):
        """ The number of attributes in the table """

        perm("r", self._access)

        self._file.lock_reading()

        leng = len(self._file.header["ATT"])

        self._file.unlock()

        return leng

    def __next__(self):
        return self.next()

    def next(self):
        tmp = None

        self._file.lock_reading()
        try:
            tmp = self._file.header["ATTBYID"][self._index][1]
        except KeyError:
            raise StopIteration
        else:
            out = self._file.header["ATT"][tmp]
            self._index += 1

            return out
        finally:
            self._file.unlock()

    def __repr__(self):
        return "<HTD%s Attributes (%s)>" % (__version__, self.location)

    @trace
    def __setitem__(self, key, v):
        if isinstance(v, Attribute) is False:
            raise ValueError("Attribute must be a Attribute instance.")

        if key is None:
            raise ValueError("Attribute name cannot be NULL.")

        self._file.lock_writing()

        header = self._file.header
        v.name = key  # Runs name check

        if v.required and v.default is None and self._file.get_length() > 0:
            raise ValueError("Table must be empty to set an attribute if it " \
                             "is required or have a default value (%s)" % key)

        new = False
        if key in header["ATT"]:
            perm("o", self._access)

            if (key in self._file.header["R1M"] or
                key in self._file.header["RM1"]):
                raise IntegrityError("Relationship(s) must be broken first")

            v._id = header["ATT"][key]._id
            old = header["ATT"][key]

            # Confirm old data
            for i in self._file.select(key):
                sanitise(unsanitise(i, {key: v}), {key: v})
        else:
            perm("a", self._access)
            v._id = len(header["ATT"])
            new = True

        header["ATT"][key] = v

        self._file.header_update(header)

        if new:
            self._file.add_attribute(key)
        else:
            # Iter through elements
            gen = self._file.modify_attribute(key, old, v)

            run = True

            try:
                nxt = next(gen)
            except StopIteration:
                run = False

            while run:
                try:
                    nxt = gen.send(sanitise(unsanitise(nxt, {key: v}), {key:
                                   v}))
                except StopIteration:
                    break

            gen.close()

        self._file.unlock()

    @trace
    def close(self):
        self.alive = False

        if self._opened:  # Close file if I opened it
            self._file.close()

        # Remove from parent
        try:
            if self in self.connection._children:
                self.connection._children.remove(self)
        except ConnectionError:
            pass

        return True

    @property
    def connection(self):
        """ Return the parent or None """

        par = self._parent()

        if par is None:
            raise ConnectionError("Parent Connection has closed.")

        return par

    @trace
    def index(self, attribute):
        """ Retrieve the index of a specific attribute """

        perm("r", self._access)

        self._file.lock_reading()

        if attribute not in self._file.header["ATT"]:
            raise KeyError("Attribute (%s) doesn't not exist." % attribute)

        i = self._file.header["ATT"][attribute]._id

        self._file.unlock()

        return i

    @trace
    def keys(self):
        """ List the keys (attribute names) in the table """

        perm("r", self._access)

        self._file.lock_reading()

        l = [a for a in self._file.header["ATT"]]

        self._file.unlock()

        return l

    @property
    def location(self):
        """ The location of the file """
        return self._file.location

    @property
    def version(self):
        """ Version of the open table """
        self._file.lock_reading()

        out = self._file.header["VER"]

        self._file.unlock()

        return out
