'''
This file is part of HTD (HayzenTech Database)

Copyright (C) 2011 - 2012 Andrew Hayzen

HTD is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

HTD is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with HTD. If not, see <http://www.gnu.org/licenses/>
'''

from .. import ConnectionError, Register
from .._base import perm, ThreadLock
from .._base.trace import register, trace
from .._objects.user import User
from weakref import ref


@register
class Auth(ThreadLock):
    """ Authentication object manages users and access for a table """
    def __init__(self, parent, location, file_obj=None):
        # Store variables
        self.alive = True
        self._parent = ref(parent)
        self.connection._children.append(self)

        self._access = parent._access  # Load access

        self._user_list = []
        self._index = 0

        ThreadLock.__init__(self)  # Start inherit

        if file_obj is None:
            htdfile = Register.resolve_fhandler(location)
            self._file = htdfile(location)  # Open file
            self._opened = True
        else:
            self._file = file_obj
            self._opened = False

    @trace
    def __contains__(self, key):
        perm("rx", self._access)

        self._file.lock_reading()

        out = key in self._file.header["USR"]

        self._file.unlock()

        return out

    @trace
    def __delitem__(self, key):
        perm("dx", self._access)

        self._file.lock_writing()

        if key not in self._file.header["USR"]:
            raise KeyError('No User, "%s", found to delete' % key)

        if len(self._file.header["USR"]) == 1:
            raise ValueError("Must be at least 2 users to delete a user")

        if self.connection.user.username == key:
            raise ValueError("Cannot delete your own user.")

        users = self._file.header["USR"]
        del users[key]

        x = sum([1 for i in users if i.access == "roadx"])
        if x < 1:
            raise ValueError("At least 1 remaining user must have full access")

        # Update header
        header = self._file.header
        header["USR"] = users
        self._file.header_update(header)

        self._file.unlock()

    @trace
    def __getitem__(self, key):
        perm("rx", self._access)

        self._file.lock_reading()

        try:
            usr = self._file.header["USR"][key]
        except KeyError:
            usr = False
            raise KeyError('No User, "%s", found' % key)

        usr.username = key  # Load username

        self._file.unlock()

        return usr

    @trace
    def __iter__(self):
        perm("rx", self._access)

        self._file.lock_reading()

        self._user_list = [u for u in self._file.header["USR"]]
        self._user_list.sort()
        self._index = 0

        self._file.unlock()

        return self

    @trace
    def __len__(self):
        perm("rx", self._access)

        self._file.lock_reading()

        out = len(self._file.header["USR"])

        self._file.unlock()

        return out

    def __next__(self):
        return self.next()

    def next(self):
        try:
            u = self._user_list[self._index]
        except IndexError:
            self._user_list = []
            self._index = 0
            raise StopIteration
        else:
            self._index += 1

            self._file.lock_reading()

            usr = self._file.header["USR"][u]

            self._file.unlock()

            return usr

    def __repr__(self):
        if self.alive is False:
            return "<HTD%s Closed Auth>" % self.version

        return "<HTD%s Auth (%s)>" % (self.version, self.location)

    @trace
    def __setitem__(self, key, v):
        if isinstance(v, User) is False:
            raise ValueError("User must be an User instance.")

        self._file.lock_writing()

        v.username = key

        if key in self._file.header["USR"]:
            perm("ox", self._access)

            if len(self._file.header["USR"]) == 1 and v.access != "roadx":
                raise ValueError("At least one User must have full access")
        else:
            perm("ax", self._access)

        header = self._file.header
        header["USR"][key] = v

        self._file.header_update(header)

        self._file.unlock()

    @trace
    def close(self):
        """ Close the cursor """
        self.alive = False  # Set closed

        if self._opened:  # Close file if I opened it
            self._file.close()

        # Remove from parent
        try:
            if self in self.connection._children:
                self.connection._children.remove(self)
        except ConnectionError:
            pass

        return True

    @property
    def connection(self):
        """ Return the parent or None """

        par = self._parent()

        if par is None:
            raise ConnectionError("Parent Connection has closed.")

        return par

    @trace
    def index(self, username):
        """ Retrieve the index of a specific username """

        perm("rx", self._access)

        self._file.lock_reading()

        if username not in self._file.header["USR"]:
            raise KeyError("User (%s) doesn't not exist." % username)

        l = [i for i in self._file.header["USR"]]
        l.sort()
        i = l.index(username)

        self._file.unlock()

        return i

    @trace
    def keys(self):
        """ List the keys (attribute names) in the table """

        perm("rx", self._access)

        self._file.lock_reading()

        l = [a for a in self._file.header["USR"]]
        l.sort()

        self._file.unlock()

        return l

    @property
    def location(self):
        """ The location of the file """
        return self._file.location

    @property
    def version(self):
        """ Version of the open table """
        self._file.lock_reading()

        out = self._file.header["VER"]

        self._file.unlock()

        return out
