'''
This file is part of HTD (HayzenTech Database)

Copyright (C) 2011 - 2012 Andrew Hayzen

HTD is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

HTD is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with HTD. If not, see <http://www.gnu.org/licenses/>

HTD Connection
'''

from .._objects.auth import Auth
from .._objects.cursor import Cursor
from .._objects.user import Authenticate, User
from .._base import ThreadLock, perm
from .._base.trace import trace, register
from .. import (__version__, AuthenticationError, ConnectionError,
                IntegrityError, Register)
from copy import copy


@register
class Connection(ThreadLock):
    """ Class that provides a connection to a table or db """
    def __init__(self, location, user=None, file_obj=None):
        ThreadLock.__init__(self)  # Load subclasses

        # Set vars
        self.alive = True
        self._children = []

        # Confirm location is str
        if not isinstance(location, str):
            raise ValueError("location (%s) must be a str object." % location)

        # Load user
        if user is None:
            user = Authenticate("guest", None)

        self.user = user.username

        # Generator for __iter__
        self._gen = None
        self._gen_cur = None

        if file_obj is None:  # Open file if not passed
            self.htdfile = Register.resolve_fhandler(location)
            file_obj = self.htdfile(location)

        self._file = file_obj

        # Load db mode
        self._isdb = False

        if self._file.header["TAB"] is False:
            self._isdb = True
        elif self._file.header["TAB"]:
            raise ConnectionError("This table is part of a db, please " \
                                  "connect to the db")

        # Load and check auth
        if user.username not in self._file.header["USR"]:
            raise AuthenticationError()

        u = self._file.header["USR"][user.username]

        if user.password != u.password:
            raise AuthenticationError()

        # Load access
        self._access = u.access

    @trace
    def __delitem__(self, name):
        """ Split table with parent - returns connection to child """

        perm("xd", self._access)

        if not self.is_db:
            raise ValueError("This feature requires a connection to a " \
                             "database")

        # Check no relationships
        cur = self[name]

        if (len(cur._file.header["R1M"]) > 0 or
            len(cur._file.header["RM1"]) > 0):
            raise IntegrityError("Relationship(s) must be broken first")

        cur = self[None]

        if name in cur:
            # Split table from parent
            data = cur[name]  # get loc
            del cur[name]
            cur.close()

            con = load_connection(data['location'])  # Connect
            cur = con[None]

            # Update header
            cur._file.lock_reading()

            header = cur._file.header
            header["TAB"] = None

            cur._file.header_update(header)

            cur._file.unlock()

            cur.close()
        else:
            cur.close()
            raise KeyError("No table to split named '%s'." % name)

    @trace
    def __getitem__(self, key):
        """ Open a cursor on the table """

        f = self._file

        if f.persistent:
            f = None

        if key is None:
            return Cursor(self, self.location, None, file_obj=f)

        # Get table data from db
        cur = Cursor(self, self.location, None, file_obj=f)
        tbd = cur[key]
        cur.close()

        return Cursor(self, tbd['location'], key)

    @trace
    def __iter__(self):
        """ If a db - return the tables """

        perm("r", self._access)

        # Check connections is DB
        if not self._isdb:
            self._gen = None
            return self

        # Open cursor, list table, close
        self._gen_cur = self[None]
        self._gen = self._gen_cur.select(["name", "location"])

        return self

    def __len__(self):
        # Check for db mode
        if self._isdb:
            cur = self[None]
            leng = len(cur)
            cur.close()
            return leng
        else:
            return 1

    def __next__(self):
        return self.next()

    def next(self):
        if not self._gen:
            raise StopIteration

        try:
            tmp = next(self._gen)
        except StopIteration:
            self._gen_cur.close()  # Close cursor
            raise StopIteration
        else:
            return tmp

    def __repr__(self):
        if self.alive is False:
            return "<HTD%s Closed Connection>" % __version__

        return ("<HTD%s Connection (%s@%s), DB Mode %s>" %
                (__version__, self.user, self.location, self.is_db))

    @trace
    def __setitem__(self, name, child):
        """ Join table (connection) with parent """

        perm('xa', self._access)

        if not self.is_db:
            return False

        if isinstance(child, self.__class__) is False:
            raise ValueError("Child must be a Connection instance")

        cur = child[None]

        # Check that child is a table
        cur._file.lock_reading()
        header = cur._file.header

        if header["TAB"] is not None:
            cur._file.unlock()
            cur.close()
            raise ValueError("Child must be a table")

        # Add location to parent
        cur = self[None]
        cur[name] = {"location": child.location}
        cur.close()

        child.close()

        cur = self[name]

        # Update child header
        cur._file.lock_writing()

        header = cur._file.header
        header["TAB"] = True

        header["USR"] = {"guest": User(None, "roadx")}  # Clear users

        cur._file.header_update(header)

        cur._file.unlock()

        cur.close()

        return self

    @property
    @trace
    def authentication(self):
        """ Create a new authentication object """

        perm("x", self._access)

        f = self._file

        if f.persistent:
            f = None

        return Auth(self, self.location, file_obj=f)

    @property
    @trace
    def children(self):
        """ Open children """

        for cur in copy(self._children):
            if cur is None:
                self._children.remove(cur)
            else:
                if not cur.alive:
                    del self._children._parent
                    self._children.remove(cur)

        return self._children

    @trace
    def close(self):
        """ Close the connection """

        for cur in copy(self.children):
            cur.close()  # Close children

        self._file.close()

        self.alive = False

        return True

    @trace
    def delete(self, children=False):
        """ Delete connection db (and tables) and disconnect """

        perm("d", self._access)

        for cur in copy(self.children):
            cur.close()  # Close cursors

        if self.is_db:
            for i in list(self):
                # Delete children
                htdfile = Register.resolve_fhandler(i["location"])
                f = htdfile(i["location"])
                f.delete_files()

        self._file.delete_files()  # Delete main file

        self.alive = False

        return True

    @property
    def is_db(self):
        """ Is the connection to a DB or table """
        return self._isdb

    def link(self, name1, name2):
        """ Link tbl1.att1 to tbl2.att2 """

        (many, att), (one, key) = self._relation_check(name1, name2)

        # Update many table
        header = many._file.header

        if att[1] not in header["RM1"]:
            header["RM1"][att[1]] = []

        if key not in header["RM1"][att[1]]:
            header["RM1"][att[1]].append(key)

        many._file.header_update(header)

        # Update one table
        header = one._file.header

        if key[1] not in header["R1M"]:
            header["R1M"][key[1]] = []

        if att not in header["R1M"][key[1]]:
            header["R1M"][key[1]].append(att)

        one._file.header_update(header)

        return True

    @property
    def location(self):
        """ The location of the file """
        return self._file.location

    def keys(self):
        """ List the table names in the db """

        if not self._isdb:
            raise ValueError("This feature requires a connection to a " \
                             "database")

        cur = self[None]

        for i in cur.keys():
            yield i

        cur.close()

    def _relation_check(self, name1, name2):
        """ Check that two attributes can form a relationship """

        if not self._isdb:
            raise ValueError("This feature requires a connection to a " \
                             "database")

        for i in (name1, name2):
            if "." not in i:
                raise ValueError("The table and attribute split by a '.' is " \
                                 "required (%s)" % i)

        # Load attributes
        tbl1, att1 = name1.rsplit(".", 1)
        tbl2, att2 = name2.rsplit(".", 1)

        for i in (tbl1, tbl2):
            if i not in self.keys():  # Check tables exist
                raise ValueError("No table found with name (%s)" % i)

        # Connect to tables
        cur1 = self[tbl1]
        cur2 = self[tbl2]

        # Check one is KEY and one is ATT
        key = None
        att = None

        for a, t, c in [(att1, tbl1, cur1), (att2, tbl2, cur2)]:
            if c.composite:
                att = (t, a)
            else:
                if a == c.identifier:
                    key = (t, a)
                else:
                    att = (t, a)

        cur1.close()
        cur2.close()

        if key is None:
            raise ValueError("One link must be to a Primary Key")
        else:
            one = self[key[0]]

        if att is None:
            raise ValueError("One link must be to an Attribute")
        else:
            many = self[att[0]]

        # Check attributes are the same
        tmp1 = one._file.header["ATT"][key[1]].to_dict()
        tmp2 = many._file.header["ATT"][att[1]].to_dict()

        del tmp1["id"]  # Strip IDs (could be different)
        del tmp2["id"]

        if tmp1 != tmp2:
            raise ValueError("Attributes to link, must be the same.")

        # Check 'many' values are in '1'
        for i in many.select(att[1]):
            if i not in one:
                raise KeyError("Value in child didn't exist in parent table " \
                               "(%s)." % i)

        return ((many, att), (one, key))

    def unlink(self, name1, name2):
        """ Unlink tbl1.att1 to tbl2.att2 """

        (many, att), (one, key) = self._relation_check(name1, name2)

        # Update many table
        header = many._file.header

        if att[1] in header["RM1"] and key in header["RM1"][att[1]]:
            header["RM1"][att[1]].remove(key)

            if len(header["RM1"][att[1]]) == 0:
                del header["RM1"][att[1]]

        many._file.header_update(header)

        # Update one table
        header = one._file.header

        if key[1] in header["R1M"] and att in header["R1M"][key[1]]:
            header["R1M"][key[1]].remove(att)

            if len(header["R1M"][key[1]]) == 0:
                del header["R1M"][key[1]]

        one._file.header_update(header)

        return True


def load_connection(*args, **kwargs):
    return Connection(*args, **kwargs)
