'''
This file is part of HTD (HayzenTech Database)

Copyright (C) 2011 - 2012 Andrew Hayzen

HTD is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

HTD is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with HTD. If not, see <http://www.gnu.org/licenses/>

HTD Cursor
'''

from .. import ConnectionError, IntegrityError, Register, __version__
from .._base import perm, Progress, ProgressItem, ThreadLock
from .._base.trace import register, trace
from .._objects.attributes import Attributes
from .._objects.record import Record
from .._workers import sanitise, unsanitise
from weakref import ref
from threading import current_thread


def integrity_check(data, header, parent):
    """ Relationship check """
    if len(header["RM1"]) > 0:
        err = False

        for att in header["RM1"]:
            for tbl, att in header["RM1"][att]:
                cur = parent[tbl]

                if data[att] not in cur and not err:
                    err = (data[att], tbl)

                cur.close()

        if err:
            raise IntegrityError("Value (%s) not in foreign table (%s)"
                                 % err)


def where_compile(where, args, atts):
    """ Compile the where criteria into eval code """

    for i in dict(args):  # Force represent
        where = where.replace("{%s}" % i, "{%s!r}" % i)

        # Sanitise data
        v = args[i]
        tmp = sanitise({v[0]: v[1]}, atts)
        args[i] = tmp[v[0]]

    where = where.format(**args)  # Load args into where
    where = compile(where, "", "eval")  # Preload eval string

    return where


@register
class Cursor(ThreadLock):
    """ Class that manages a cursor on a table """
    def __init__(self, parent, location, table, header=None, file_obj=None):
        # Store variables
        self.alive = True
        self.table = table

        self._i = 0  # index for __iter__

        self._parent = ref(parent)
        self.connection._children.append(self)

        self._access = parent._access

        ThreadLock.__init__(self)  # Start inherit

        # Open file
        if file_obj is None:
            htdfile = Register.resolve_fhandler(location)
            self._file = htdfile(location, header)
            self._opened = True
        else:
            self._file = file_obj
            self._opened = False

    @trace
    def __contains__(self, key):
        perm("r", self._access)

        self._file.lock_reading()  # Lock file

        # Filter primarykey and get line
        if len(self._file.header["KEY"]) == 1:
            key = [key]

        key = sanitise(dict(zip(self._file.header["KEY"], key)),
                       self._file.header["ATT"])

        line = self._file.get_line(key)

        self._file.unlock()  # Unlock

        return line[0] is not False  # Return True if key exists

    @trace
    def __delitem__(self, key):
        perm("d", self._access)

        self._file.lock_writing()

        if len(self._file.header["R1M"]) > 0:
            raise IntegrityError("Relationship must be broken first")

        if isinstance(key, slice):  # delete a slice
            start = key.start
            if start is None:
                start = 0

            stop = key.stop
            if stop is None:
                stop = len(self._file)

            if start >= stop:
                raise ValueError("Stop index must be greater than start.")

            self._file.delete_region(start=start, stop=stop)

            self._file.unlock()
        else:
            # San data
            if isinstance(key, (list, tuple)) is False:
                key = [key]

            key = sanitise(dict(zip(self._file.header["KEY"], key)),
                           self._file.header["ATT"])

            # Check if record exists and get line

            pos, data, nline = self._file.get_line(key)
            if pos is False:
                raise KeyError('No record, %s, was found to delete.' % (key))

            del self._file[pos]  # Delete i from file

            self._file.unlock()

    @trace
    def __getitem__(self, key):
        perm("r", self._access)

        if isinstance(key, slice):
            return self.select(start=key.start or 0, stop=key.stop or False)

        self._file.lock_reading()  # Lock file

        # Convert key
        if isinstance(key, (list, tuple)) is False:
            key = [key]

        key = sanitise(dict(zip(self._file.header["KEY"], key)),
                       self._file.header["ATT"])

        pos, data, nline = self._file.get_line(key)  # Check key exists

        if pos is False:
            raise KeyError("".join(['No key, ', str(key), ', found']))

        data = unsanitise(data, self._file.header["ATT"])

        self._file.unlock()  # Unlock

        return Record(key, data, self.connection, self.table,
                      self._file.header["KEY"])

    @trace
    def __iter__(self):
        """ If a db - return the tables """
        perm("r", self._access)

        self._i = 0

        return self

    def __next__(self):
        return self.next()

    def next(self):
        self._file.lock_reading()

        data = self._file[self._i]

        if data is False:
            self._i = 0
            self._file.unlock()
            raise StopIteration

        data = unsanitise(data, self._file.header["ATT"])

        self._i += 1

        self._file.unlock()

        return Record(None, data, self.connection, self.table,
                      self._file.header["KEY"])

    @trace
    def __len__(self):
        """ The number of records in the table """

        self._file.lock_reading()

        length = len(self._file)  # Get length

        self._file.unlock()

        return length

    def __repr__(self):
        if self.alive is False:
            return "<HTD%s Closed Cursor>" % __version__

        return "<HTD%s Cursor (%s)>" % (__version__, self.location)

    @trace
    def __setitem__(self, key, data):
        self._file.lock_writing()

        # Filter primarykey
        if isinstance(key, (list, tuple)) is False:
            key = [key]

        key = dict(zip(self._file.header["KEY"], key))

        # Add identifier to data
        for att in self._file.header["KEY"]:
            data[att] = key[att]

        integrity_check(data, self._file.header, self.connection)

        data = sanitise(data, self._file.header["ATT"])  # Run filters

        # Check if record exists and get current att
        pos, datatmp, nline = self._file.get_line(data)

        # If record exists get current record
        if pos is False:
            datatmp = {}

        # Confirm data
        for att in self._file.header["ATT"]:
            if att not in data:
                if att in datatmp:
                    data[att] = datatmp[att]
                else:
                    data[att] = None

        # Write to file
        if pos is not False:
            perm("o", self._access)

            if len(self._file.header["R1M"]) > 0:
                raise IntegrityError("Relationship must be broken first")

            self._file[pos] = data
        else:
            perm("a", self._access)
            self._file.insert_line(nline, data)

        self._file.unlock()

    @property
    def attributes(self):
        """ Manage the attribute settings """
        if self._opened:
            htdfile = None
        else:
            htdfile = self._file

        return Attributes(self.connection, self.location, htdfile)

    @trace
    def close(self):
        """ Close the cursor """
        self.alive = False  # Set closed

        if self._opened:  # Close file if I opened it
            self._file.close()

        # Remove from parent
        try:
            if self in self.connection._children:
                self.connection._children.remove(self)
        except ConnectionError:
            pass

        return True

    @property
    def composite(self):
        """ Is the table using a composite key """

        return isinstance(self.identifier, (list, tuple))

    @property
    def connection(self):
        """ Return the parent or None """

        par = self._parent()

        if par is None:
            raise ConnectionError("Parent Connection has closed.")

        return par

    @trace
    def delete(self, where, args):
        """ Delete items in the table which meet a criteria """

        perm("d", self._access)

        self._file.lock_writing()

        if len(self._file.header["R1M"]) > 0:
            raise IntegrityError("Relationship must be broken first")

        if where is not None:  # Load where
            where = where_compile(where, args, self._file.header["ATT"])

        x = 0
        c = 0  # Count number of deleted items
        size = len(self._file)

        while x < size:
            record = self._file[x]  # Read record

            if where is not None:  # Check where clause
                if not eval(where, dict(record)):
                    x += 1  # Next record
                    continue

            del self._file[x]

            size -= 1  # Size decreases as items are deleted
            c += 1

        self._file.unlock()

        return c

    @property
    @trace
    def identifier(self):
        """ The identifier for the table """

        perm('r', self._access)

        self._file.lock_reading()

        key = self._file.header["KEY"][:]

        if len(key) == 1:
            key = key[0]

        self._file.unlock()

        return key

    @identifier.setter
    @trace
    def identifier(self, key):
        """ Set the identifier for the table """

        perm("o", self._access)

        self._file.lock_writing()

        # Confirm data
        if isinstance(key, str):
            key = [key]

        if isinstance(key, (list, tuple)) is False:
            raise ValueError("Identifier must be a list or list of strings")

        if len(self._file) > 0:
            raise ValueError("The table must be empty before changing the" \
                             " identifier.")

        header = self._file.header
        header["KEY"] = key

        self._file.header_update(header)

        self._file.unlock()

    @trace
    def index(self, key):
        """ Retrieve the index of a key """

        perm("r", self._access)

        self._file.lock_reading()  # Lock file

        # Convert key
        if isinstance(key, (list, tuple)) is False:
            key = [key]

        key = sanitise(dict(zip(self._file.header["KEY"], key)),
                       self._file.header["ATT"])

        # Check key exists
        pos, data, nline = self._file.get_line(key)

        if pos is False:
            raise KeyError("".join(['No key, ', str(key), ', found']))

        self._file.unlock()  # Unlock

        return pos

    @trace
    def insert(self, generator):
        """ Insert a generator into the table """

        perm("a", self._access)

        self._file.lock_writing()

        gen = self._file.insert()  # Load file generator

        next(gen)

        for record in generator:  # Loop and send
            integrity_check(record, self._file.header, self.connection)

            key = [record[a] for a in self._file.header["KEY"]]
            key = sanitise(dict(zip(self._file.header["KEY"], key)),
                           self._file.header["ATT"])

            gen.send((key, sanitise(record, self._file.header["ATT"])))

        self._file.unlock()

    @trace
    def keys(self):
        """ List the keys in the table """

        perm("r", self._access)

        return self.select(self._file.header["KEY"])

    @property
    def location(self):
        """ The location of the file """
        return self._file.location

    @trace
    def select(self, attribute=None, start=0, stop=False, where=None, args={}):
        """ List selected attribute(s) """

        perm("r", self._access)

        self._file.lock_reading()

        if attribute is None:  # Select all atts
            attbyid = self._file.header["ATTBYID"]
            attribute = [attbyid[i][1] for i in attbyid]

        if isinstance(attribute, str):
            attribute = [attribute]

        if not isinstance(attribute, (list, tuple)):
            raise ValueError("Attribute must be a list, tuple or str.")

        numatts = len(attribute)  # Count atts

        # Check atts exist
        if sum([1 for att in attribute if att not in
                self._file.header["ATT"]]) > 0:
            raise KeyError('No find attribute, %s, to list.' % attribute)

        if stop is False:  # load stop
            stop = len(self._file)

        if start > stop:
            raise ValueError("Stop index must be greater than start.")

        if where is not None:  # Load where
            where = where_compile(where, args, self._file.header["ATT"])

        gen = self._file.select(attribute, start, stop)  # load generator

        # Load progress
        calling = "Calling htd.Cursor.select"
        prog = ProgressItem(current_thread(), stop - start, calling)
        Progress.add(str(id(prog)), prog)

        for i in gen:
            if where is not None:  # Check where clause
                if not eval(where, dict(i)):
                    continue

            tmp = unsanitise(i, self._file.header["ATT"])

            self._file.unlock()

            prog.count()

            if numatts == 1:
                yield tmp[attribute[0]]
            else:
                yield tmp

            self._file.lock_reading()

        Progress.remove(str(id(prog)))

        self._file.unlock()

    @trace
    def update(self, where, args):
        """ Update a generator into the table """

        perm("o", self._access)

        self._file.lock_writing()

        if len(self._file.header["R1M"]) > 0:
            raise IntegrityError("Relationship must be broken first")

        if where is not None:  # Load where
            where = where_compile(where, args, self._file.header["ATT"])

        x = 0

        while x < len(self._file):
            record = self._file[x]  # Read record

            if where is not None:  # Check where clause
                if not eval(where, dict(record)):
                    x += 1  # Next record
                    continue

            record = unsanitise(record, self._file.header["ATT"])
            record = (yield record)  # Send/Recieve new record

            integrity_check(record, self._file.header, self.connection)

            self._file[x] = sanitise(record, self._file.header["ATT"])

            x += 1

            yield x

        self._file.unlock()

    @property
    @trace
    def version(self):
        """ Version of the open table """
        perm("r", self._access)

        self._file.lock_reading()

        out = self._file.header["VER"]

        self._file.unlock()

        return out
