'''
This file is part of HTD (HayzenTech Database)

Copyright (C) 2011 - 2012 Andrew Hayzen

HTD is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

HTD is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with HTD. If not, see <http://www.gnu.org/licenses/>

HTD Record
'''

from .. import ConnectionError
from weakref import ref


class Record(object):
    """ Class that provides read-only record and foreign key support """
    def __init__(self, key, data={}, con=None, table=None, identifier=[]):
        # Load variables
        if con is not None:
            self._con = ref(con)
        else:
            self._con = None

        self.data = data
        self.key = key
        self.table = table
        self._index = 0

    def __contains__(self, key):
        return key in self.data

    def __getitem__(self, key):
        if "." not in key:  # Check if foreign key lookup needed
            return self.data[key]  # Return key from data

        # Split table and attribute
        key = key.split(".")
        table = key[-1]
        key = key[:-1]

        # Check the object exists
        try:
            cur = self.connection[table]  # Open cursor
        except KeyError:
            raise KeyError("No table found (%s)" % table)

        # Load relationships
        r1m = cur._file.header["R1M"]
        rm1 = cur._file.header["RM1"]
        r1m_link = []
        rm1_link = []

        # Find relationships that match key.table
        for att in r1m:
            for tbl, att in r1m[att]:
                if tbl == self.table and att in key:
                    r1m_link.append((tbl, att))

        for att in rm1:
            for tbl, att in rm1[att]:
                if tbl == self.table and att in key:
                    rm1_link.append((tbl, att))

        # Check that relationships are valid
        if len(r1m_link) == 0 and len(rm1_link) == 0:
            raise ValueError("Attribute (%s) has no relation to table (%s)." %
                             (key, table))

        if len(r1m_link) > 0 and len(rm1_link) > 0:
            raise ValueError("Attribute (%s) has too many relationships to" \
                             " table (%s)." % (key, table))

        if len(r1m_link) > 0:  # For 1-m relationships
            tbl, att = r1m_link[0]  # Pick first relation

            data = cur[self.data[att]]

        if len(rm1_link) > 0:  # For m-1 relationships
            tbl, att = rm1_link[0]  # Pick first relation

            data = list(cur.select(where="%s=={val}" % att,
                                   args=dict(val=(att, self.data[att]))))

        cur.close()  # Close cursor

        return data  # Return data

    def __iter__(self):
        self._index = 0
        return self

    def __len__(self):
        """ Return number fields """
        return len(self.data)

    def __next__(self):
        return self.next()

    def next(self):
        try:
            d = list(self.data.keys())[self._index]
        except IndexError:
            self._index = 0
            raise StopIteration
        else:
            self._index += 1
            return d

    def __repr__(self):
        return str(self.data)

    def __setitem__(self, key, value):
        self.data[key] = value

    @property
    def connection(self):
        """ Return the parent or None """

        if self._con is None:
            return None

        par = self._con()

        if par is None:
            raise ConnectionError("Parent Connection has closed.")

        return par
