'''
This file is part of HTD (HayzenTech Database)

Copyright (C) 2011 - 2012 Andrew Hayzen

HTD is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

HTD is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with HTD. If not, see <http://www.gnu.org/licenses/>

Data Filter to import/export raw data
Also loads default file handlers
get_line alog and locking classes
'''

from .. import AuthenticationError
from .._base import File, convert_location
import os
from random import choice
from re import sub
from struct import unpack, pack
from sys import platform
from time import time

# Load fsync
if platform == 'linux2':
    fsync = os.fdatasync
else:
    fsync = os.fsync


def sanitise(data, attributes):
    """ Sanitise data so that it can be stored in the db """

    if data is None:
        return None

    # Check data attributes exist
    atts = [a for a in data if a not in attributes]

    if len(atts) > 0:
        raise AttributeError('Failed to find sanitise settings for ' \
                                 'attribute %s' % atts)

    for att in attributes:
        if att in data:  # Load data
            tmp = data[att]
        else:
            tmp = None

        attr = attributes[att]  # load attribute

        if tmp is None:
            if attr._default is not None:
                tmp = attr._default  # Load default value
                continue
            elif attr._required:  # Check required
                raise ValueError("Required attribute (%s) was NULL" % att)
            else:
                continue

        typed = attr._dtype  # Load dtype

        if typed.type_check(tmp) is False:  # Type check
            name = ('%s, failed to  match preset pattern for %s. (%s)' %
                    (att, typed, tmp))

            raise ValueError(name)

        if typed.can_range and attr._range is not False:  # Range check
            if not (attr._range[1] < tmp < attr._range[0]):
                raise ValueError('"%s" failed to match range for "%s". (%s)' %
                                 (att, typed, tmp))

        if attr._limit_list[0] and tmp not in attr._limit_list[1]:  # List
            raise KeyError('"%s" failed to match an item in list.' % att)

        tmp = typed.sanitise(tmp)  # Convert data

        if len(tmp) > attr._size:  # Length Check
            raise ValueError('String length check failed for "%s". (%s)' %
                             (att, tmp))

        data[att] = tmp  # Update data

    return data


def unsanitise(data, attributes):
    """ Unsantise data so that it can be used by the client """
    for att in data:
        dtype = attributes[att]._dtype
        patt = attributes[att]._re

        if data[att] is not None:
            if not not patt:  # Run reg expr
                data[att] = sub(patt, b"", data[att])

            data[att] = dtype.unsanitise(data[att])

    return data


class GetLineAlgorithm(object):
    def get_line(self, key, new=False):
        """ Find line for key """
        # Set up atts, key
        atts = self.header["KEY"]
        primarykey = b"".join([key[att] for att in atts])

        # Calculate upper and lower  Will take log(2, numrecords)
        size = len(self)
        upper = size
        lower = 0

        # Find line
        datal = None
        datap = []
        cline = upper // 2
        nline = cline

        if size == 0:  # Check table not empty
            cline = False
            nline = 0
            return [cline, datal, nline]

        while 1:
            datal = self[cline]  # Load cline

            if datal is False:  # Fix for end of file
                nline = cline
                cline = False
                break

            datap = b"".join([datal[a] for a in atts])  # Build data

            if datap == primarykey:  # Check data not primarykey
                nline = False
                break

            comp = (datap > primarykey) - (datap < primarykey)  # Compare

            # Calc movement
            # upper 10 lower 0 cline 5
            if comp > 0:
                upper = cline
                if (upper - lower) == 1:  # upper 1 lower 0 cline 1 => 0
                    cline = lower
                    continue
                else:  # upper 5 lower 0 cline 5 => 2.5
                    cline = lower + int((upper - lower) // 2)
            elif comp < 0:
                if (upper - lower) == 1:  # upper 1 lower 0 cline 0 => 1
                    cline = upper
                    lower = upper
                    continue
                else:  # upper 10 lower 0 cline 5 => 7.5
                    lower = cline
                    cline = lower + int((upper - lower) // 2)

            if upper == lower:
                nline = cline
                cline = False
                break

        return [cline, datal, nline]

    def insert(self):
        while 1:
            key, data = (yield)

            pos, datal, nline = self.get_line(key, new=True)

            if pos is False:
                self.insert_line(nline, data)
            else:
                raise KeyError("Key already exists (%s)" % key)


class LockFile():
    def __init__(self, location):
        self.__location = convert_location(location, ".~lock_")

        if not os.path.exists(self.__location):
            f = open(self.__location, "wb")
            f.close()

        self.__file = File(self.__location)

        while 1:
            try:
                self.__file._lock_reading()
                break
            except IOError:
                pass

        self._files = [self.__file]

        self._confirm_funcs = []

    @property
    def locked(self):
        """ Return the locks on the cursor """
        locked = True

        try:
            self.__file._lock_writing()
            locked = False
        except IOError:
            pass
        finally:
            self.__file._lock_reading()

        return locked

    def lock_reading(self):
        self._acquire()

        while 1:
            try:
                self._file._lock_reading()
                break
            except IOError:
                pass

        for i in self._files:
            i.flush()  # Flush files

        for i in self._confirm_funcs:
            i()

        return

    def lock_writing(self):
        self._acquire()

        while 1:
            try:
                self._file._lock_writing()
                break
            except IOError:
                pass

        for i in self._files:
            i.flush()  # Flush files

        for i in self._confirm_funcs:
            i()

        return

    def unlock(self):
        for i in self._files:
            i.flush()  # Flush files

        if self.force_sync and self._file.lock_type:
            for i in self._files:
                fsync(i)  # sync file if lock_writing

        try:
            self._file._unlock()
        except IOError:
            pass

        if self._thread_locked:
            self._release()

        return

# Import default file handlers
from . import file_var
from . import file_fix
