'''
This file is part of HTD (HayzenTech Database)

Copyright (C) 2011 - 2012 Andrew Hayzen

HTD is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

HTD is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with HTD. If not, see <http://www.gnu.org/licenses/>

HTD_FILE_FIXED_WIDTH
'''

from .. import Config, Register
from .._base import File, ThreadLock, convert_location, location_is_file
from .._workers import GetLineAlgorithm, LockFile
from .._workers.header import Header
import os


class FIXED_WIDTH_FILE(ThreadLock, Header, GetLineAlgorithm, LockFile):
    file_format = b"\x05"
    persistent = True

    def __init__(self, location, header=None):
        # Load settings
        self.force_sync = Config.force_sync
        self._rec_width = 0

        # Load classes
        ThreadLock.__init__(self)
        Header.__init__(self)
        GetLineAlgorithm.__init__(self)

        # Check for file
        if not os.path.exists(location):
            if header is None:
                raise IOError("No HTD file found (%s)" % location)

            f = open(location, "wb")
            f.close()

            self._file = File(location)
            self._Header__update(header, True)
        else:
            if header is not None:
                raise IOError('File already exists (%s)' % location)

            self._file = File(location)

        self.header_update()  # Load header

        # Set variables
        self.location = location
        self.escape_char = b"\x00"

        LockFile.__init__(self, location)
        self._files.append(self._file)
        self._confirm_funcs.append(self.header_confirm)

    def __delitem__(self, key):
        # Delete record
        pos = (self._rec_width * key) + self.header["SIZ"]

        del self._file[pos:pos + self._rec_width]  # Delete from file

    def __getitem__(self, key):
        pos = (self._rec_width * key) + self.header["SIZ"]  # Find pos

        # Loop through attributes
        output = {}
        atts = self.header["ATTBYID"]

        self._file.seek(pos)

        r = self._file.read

        # Read atts
        for att in atts:
            data = r(atts[att][0] + 1)

            if data == b"":
                return False

            data = data.rsplit(b".", 1)

            if len(data) == 1:
                output[atts[att][1]] = None
            else:
                output[atts[att][1]] = data[0]

        return output

    def __len__(self):
        pos = len(self._file) - self.header["SIZ"]
        return pos // self._rec_width

    def __setitem__(self, key, value):
        # Write record
        # Record MUST exist
        pos = (self._rec_width * key) + self.header["SIZ"]

        # Loop through attributes
        string = []
        atts = self.header["ATTBYID"]

        for att in atts:
            if value[atts[att][1]] is None:
                d = b""
            else:
                d = b"".join([value[atts[att][1]], b"."])

            string.append(d.ljust(atts[att][0] + 1, self.escape_char))

        string = b"".join(string)

        self._file[pos:pos + self._rec_width] = string

    def _modify_attribute(self, attribute, mode):
        if mode == 1:  # Calc old length
            pos = len(self._file) - self.header["SIZ"]
            leng = pos // (self._rec_width -
                           self.header["ATT"][attribute].size)
        elif mode == -1:
            leng = len(self)

        if leng == 0:
            return True

        atts = self.header["ATTBYID"]
        self._file.seek(self.header["SIZ"])

        x = 0
        pos = self.header["SIZ"]
        while x < leng:
            for a in atts:
                size = atts[a][0] + 1
                if atts[a][1] == attribute:
                    if mode == 1:
                        self._file[pos:pos] = self.escape_char * size
                        pos += size
                    elif mode == -1:
                        del self._file[pos:pos + size]
                else:
                    pos += atts[a][0] + 1
            x += 1

        return True

    def add_attribute(self, attribute):
        return self._modify_attribute(attribute, 1)

    def close(self):
        """ Close file and sub classes """
        # Remove self from cache file
        self._file._lock_writing()

        locked = self.locked

        # Close lock file
        self._LockFile__file.close()

        # Check for other locks
        if not locked:
            try:
                os.remove(self._LockFile__location)
            except IOError:
                pass

        # Close file
        self._file._unlock()
        self._file.close()

        return True

    def delete_attribute(self, attribute):
        return self._modify_attribute(attribute, -1)

    def delete_region(self, start=None, stop=None):
        spos = (self._rec_width * start) + self.header["SIZ"]
        epos = (self._rec_width * stop) + self.header["SIZ"]

        del self._file[spos:epos]

        return True

    def delete_files(self):
        # Close files
        for f in self._files:
            f.close()

        # Delete table
        os.unlink(self.location)

        lock_loc = convert_location(self.location, ".~lock_")
        if os.path.exists(lock_loc):
            os.unlink(lock_loc)

        return True

    def header_update(self, header=None):
        if header is not None:
            self._Header__update(header)
        else:
            self._Header__load()

        self._rec_width = sum([self.header["ATT"][i].size + 1 for i in
                               self.header["ATT"]])
        return

    def insert_line(self, key, data):
        # Write record
        # Record MUST NOT exist
        pos = (self._rec_width * key) + self.header["SIZ"]

        # Loop through attributes
        string = []
        atts = self.header["ATTBYID"]

        for att in atts:
            if data[atts[att][1]] is None:
                d = b""
            else:
                d = b"".join([data[atts[att][1]], b"."])

            string.append(d.ljust(atts[att][0] + 1, self.escape_char))

        string = b"".join(string)

        self._file[pos:pos] = string

        return

    def modify_attribute(self, attribute, old, new):
        nsize = new.size + 1
        osize = old.size + 1

        # Calc old length
        pos = len(self._file) - self.header["SIZ"]
        leng = pos // (self._rec_width - (nsize - osize))

        if leng == 0:
            raise StopIteration

        atts = self.header["ATTBYID"]
        self._file.seek(self.header["SIZ"])

        x = 0
        pos = self.header["SIZ"]

        while x < leng:
            for a in atts:
                if atts[a][1] == attribute:
                    raw = self._file[pos:pos + osize]

                    if raw == b"":  # break if end of file
                        raise StopIteration

                    raw = raw.rsplit(b".", 1)

                    if len(raw) == 1:
                        raw = None
                    else:
                        raw = raw[0]

                    tmp = (yield {attribute: raw})

                    if tmp[attribute] is None:
                        d = b""
                    else:
                        d = b"".join([tmp[attribute], b"."])

                    output = d.ljust(nsize, self.escape_char)

                    self._file[pos:pos + osize] = output

                    pos += nsize
                else:
                    pos += atts[a][0] + 1
            x += 1

        raise StopIteration

    def select(self, attribute, start, stop):
        pos = start
        atts = self.header["ATTBYID"]
        s = self._file.seek
        r = self._file.read

        while pos < stop:
            output = {}
            s((self._rec_width * pos) + self.header["SIZ"])

            for att in atts:   # Read atts
                if atts[att][1] not in attribute:
                    s(atts[att][0] + 1, 1)
                    continue

                d = r(atts[att][0] + 1)

                if d == b"":
                    raise StopIteration

                d = d.rsplit(b".", 1)

                if len(d) == 1:
                    d = None
                else:
                    d = d[0]

                output[atts[att][1]] = d

            pos += 1
            yield output

        raise StopIteration


Register.register_fhandler(FIXED_WIDTH_FILE, location_is_file)
