'''
This file is part of HTD (HayzenTech Database)

Copyright (C) 2011 - 2012 Andrew Hayzen

HTD is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

HTD is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with HTD. If not, see <http://www.gnu.org/licenses/>

HTD_FILE_VARIABLE_WIDTH
'''

from .. import Config, Register
from .._base import File, ThreadLock, convert_location, location_is_file
from .._workers import GetLineAlgorithm, LockFile
from .._workers.header import Header

import os
from struct import error, unpack, pack
from time import time, sleep

cache_size = 50


class Index():
    def __init__(self, location):
        # Set vars
        self.__cache = []
        self.__cache_size = 53
        self.__location = convert_location(location, ".~index_")
        self.__file = None
        self.__tms = (0, 0)

        # Create and connect to index
        if not os.path.exists(self.__location):
            self.__create_file(self._file, self.header["SIZ"],
                               len(self.header["ATT"]))
        else:
            self.__file = File(self.__location)
            self.__create_cache()

    def __close(self):
        """ Disconnect from index """
        if self.__file is not None:
            self.__file.close()
            self.__file = None

        self.__cache = []
        self.__cache_size = 50

        return

    def __create_cache(self, change=True):
        """ Create cache from connected file """
        # Calculate Cache Size
        length = self.__get_length()
        self.__cache_size = int((length * 0.01033) + 52.77)

        # Work out how many items
        length = length // self.__cache_size

        # Set vars
        x = 0
        output = []
        txt = "Q" * self.__cache_size
        siz = self.__cache_size * 8

        # Seek
        self.__file.seek(9)

        # Read file
        while x < length:
            tmp = self.__file.read(siz)
            output.append(sum(unpack(txt, tmp)))
            x += 1

        # Set cache
        self.__cache = output

        if change:
            self.__create_tms()

        return True

    def __create_file(self, fileobj, pos, numatts):
        """ Create index file from fileobj """
        # Create file
        if not os.path.exists(self.__location):
            f = open(self.__location, "wb+")
            f.close()

        if self.__file is None:
            self.__file = File(self.__location)

        self.__clear()  # Clear cache

        # Get length of data and seek to start
        fileobj.seek(0, 2)
        end = fileobj.tell()
        fileobj.seek(pos)

        self.__file.seek(0)
        self.__file.write(b"\x00" * 9)

        # Loop through data
        while pos < end:
            length = 0
            x = 0

            while x < numatts:
                d = fileobj.read(4)
                wid = unpack("I", d)[0]

                if wid == 4294967295:
                    wid = 0

                fileobj.seek(wid, 1)

                wid += 4
                pos += wid
                length += wid
                x += 1

            self.__file.write(pack("Q", length))

        # Create cache
        self.__create_cache()
        return True

    def __create_tms(self):
        # Load timestamp
        while 1:
            timestamp = int(time() * 10 ** 6)
            if timestamp != self.__tms[0]:
                time_id = 1
                break
            elif self.__tms[1] == 255:
                sleep(0.1)
                continue
            else:
                time_id = self.__tms[1]
                break

        self.__tms = (timestamp, time_id)
        self.__file[0:9] = pack("QB", *self.__tms)

        return

    def index_confirm(self):
        """ Check for changes in the header """
        self._file.seek(0)

        timestamp, time_id = unpack("QB", self._file.read(9))
        otimestamp, otime_id = self.__tms  # Load old times

        if otime_id != time_id or otimestamp != timestamp:
            self.__create_cache()

        return

    def __clear(self):
        """ Clear index """
        if self.__file is not None:
            self.__file.seek(9)
            self.__file.truncate()

        # Create cache
        self.__create_cache()
        return True

    def __clear_slice(self, start, stop):
        """ Clear index """
        if self.__file is not None:
            # Convert -1 to max
            if stop == -1:
                stop = self.__get_length()

            # Clear slice
            del self.__file[start * 8 + 9:stop * 8 + 9]

        # Create cache
        self.__create_cache()
        return True

    def __delete(self):
        """ Delete index file """
        if os.path.exists(self.__location) is False:
            return False

        self.__cache = []
        os.unlink(self.__location)
        return True

    def __delete_key(self, key):
        """ Delete a record from index """
        key *= 8
        del self.__file[key + 9:key + 17]

        self.__create_cache()
        return True

    def __get_length(self):
        """ Get length (number of records) of index """
        return (len(self.__file) - 9) // 8

    def __get_position(self, key):
        """ Get position in db of key """
        # Convert -1 to max
        if key == -1:
            key = self.__get_length()

        # Load cache
        cache = key // self.__cache_size
        if cache == 0:
            length = 0
        else:
            length = sum(self.__cache[0:cache])

        key = key % self.__cache_size

        self.__file.seek(cache * (self.__cache_size * 8) + 9)

        # Read width
        data = self.__file.read(key * 8)

        if key == 1:
            length += unpack("Q", data)[0]
        elif key == 0:
            pass
        else:
            try:
                length += sum(unpack("Q" * key, data))
            except error:
                return False

        return length + self.header["SIZ"]

    def __get_width(self, key):
        """ Get length of record """
        self.__file.seek(8 * key + 9)
        data = self.__file.read(8)

        if len(data) == 0:
            return 0
        else:
            return unpack("Q", data)[0]

    def __insert(self, pos, value):
        """ Insert record into position """
        data = pack("Q", value)
        pos *= 8
        self.__file[pos + 9:pos + 9] = data

        self.__create_cache()
        return True

    def __set(self, key, value):
        """ Change record width to value """
        key *= 8
        data = pack("Q", value)
        self.__file[key + 9:key + 17] = data

        self.__create_cache()
        return True


class VARIABLE_WIDTH_FILE(ThreadLock, Header, Index, GetLineAlgorithm,
                          LockFile):
    file_format = b"\x01"
    persistent = True

    def __init__(self, location, header=None):
        # Load settings
        self.force_sync = Config.force_sync

        # Load classes
        ThreadLock.__init__(self)
        Header.__init__(self)
        GetLineAlgorithm.__init__(self)

        # Check for file
        if not os.path.exists(location):
            if header is None:
                raise IOError("No HTD file found (%s)" % location)

            f = open(location, "wb")
            f.close()

            self._file = File(location)
            self._Header__update(header, True)
            new = True
        else:
            if header is not None:
                raise IOError('File already exists (%s)' % location)

            self._file = File(location)
            new = False

        self.header_update()  # Load header

        # Set variables
        self.location = location

        LockFile.__init__(self, location)
        self._files.append(self._file)
        self._confirm_funcs.append(self.header_confirm)

        Index.__init__(self, location)
        self._files.append(self._Index__file)
        self._confirm_funcs.append(self.index_confirm)

        if new:
            self._Index__clear()

    def __delitem__(self, key):
        # Delete record

        # Find pos
        pos = self._Index__get_position(key)
        if pos is False:
            raise IndexError("No record found to delete (%s)" % key)

        wid = self._Index__get_width(key)  # Get width

        del self._file[pos:pos + wid]  # Delete from file

        self._Index__delete_key(key)  # Update index and cache

    def __len__(self):
        return self._Index__get_length()

    def __getitem__(self, key):
        # Find pos
        pos = self._Index__get_position(key)

        if pos is False:
            return False

        # Loop through attributes
        output = {}
        atts = self.header["ATTBYID"]

        self._file.seek(pos)

        r = self._file.read

        # Read atts
        for att in atts:
            try:
                wid = unpack("I", r(4))[0]
            except error:
                output = False
                break

            if wid == 4294967295:
                output[atts[att][1]] = None
            else:
                output[atts[att][1]] = r(wid)

        return output

    def __setitem__(self, key, value):
        # Write record
        # Record MUST exist
        pos = self._Index__get_position(key)

        if pos is False:
            raise IndexError("No record found to set (%s)" % key)

        # Get width
        owid = self._Index__get_width(key)

        # Loop through attributes
        string = []
        atts = self.header["ATTBYID"]

        for att in atts:
            val = value[atts[att][1]]
            if val is None:
                string.append(b"\xff" * 4)
            else:
                wid = pack("I", len(val))
                string.extend([wid, val])

        string = b"".join(string)

        self._file[pos:pos + owid] = string

        # Update index
        wid = len(string)
        if owid != wid:
            self._Index__set(key, wid)

    def _modify_attribute(self, attribute, mode):
        leng = len(self)
        if leng == 0:
            return True

        attributes = self.header["ATTBYID"]
        self._file.seek(self.header["SIZ"])

        x = 0
        pos = self.header["SIZ"]

        while x < leng:
            for a in attributes:
                if attributes[a][1] == attribute:
                    if mode == 1:
                        self._file[pos:pos] = b"\xff" * 4
                        pos += 4
                    elif mode == -1:
                        wid = unpack("I", self._file.read(4))[0]
                        if wid == 4294967295:
                            wid = 0

                        del self._file[pos:pos + wid + 4]

                    self._file.seek(pos)
                else:
                    wid = unpack("I", self._file.read(4))[0]
                    self._file.seek(wid, 1)
                    pos += wid + 4

            x += 1

        num_atts = len(self.header["ATT"])

        if mode == -1:
            num_atts -= 1

        self._Index__create_file(self._file, self.header["SIZ"], num_atts)

        return True

    def add_attribute(self, attribute):
        return self._modify_attribute(attribute, 1)

    def close(self):
        """ Close file and sub classes """
        # Remove self from cache file
        self._file._lock_writing()

        self._Index__close()  # Close index

        locked = self.locked

        # Close lock file
        self._LockFile__file.close()

        # Check for other locks
        if not locked:
            try:
                os.remove(self._LockFile__location)
            except IOError:
                pass

        # Close file
        self._file._unlock()
        self._file.close()

        return True

    def delete_attribute(self, attribute):
        return self._modify_attribute(attribute, -1)

    def delete_files(self):
        # Close files
        for f in self._files:
            f.close()

        # Delete table
        os.unlink(self.location)

        for i in [".~lock_", ".~index_"]:
            loc = convert_location(self.location, i)
            if os.path.exists(loc):
                os.unlink(loc)

        return True

    def delete_region(self, start=None, stop=None):
        spos = self._Index__get_position(start)
        epos = self._Index__get_position(stop)

        del self._file[spos:epos]

        self._Index__clear_slice(start, stop)

        return True

    def header_update(self, header=None):
        if header is not None:
            self._Header__update(header)
        else:
            self._Header__load()

        return

    def insert_line(self, key, data):
        # Write record (Record MUST NOT exist)
        pos = self._Index__get_position(key)

        if pos is False:
            return False

        # Loop through attributes
        string = []
        atts = self.header["ATTBYID"]

        for att in atts:
            val = data[atts[att][1]]
            if val is None:
                string.append(b"\xff" * 4)
            else:
                wid = pack("I", len(val))
                string.extend([wid, val])

        string = b"".join(string)

        wid = len(string)
        self._file[pos:pos] = string

        self._Index__insert(key, wid)  # Update index

        return

    def modify_attribute(self, attribute, old, new):
        leng = len(self)

        if leng == 0:
            raise StopIteration

        atts = self.header["ATTBYID"]
        self._file.seek(self.header["SIZ"])
        num_atts = len(self.header["ATT"])

        x = 0
        pos = self.header["SIZ"]

        while x < leng:
            for a in atts:
                self._file.seek(pos)
                if atts[a][1] == attribute:
                    wid = unpack("I", self._file.read(4))[0]  # read old

                    if wid == 4294967295:
                        wid = 0
                        raw = None
                    else:
                        raw = self._file[pos + 4:pos + wid + 4]

                    if raw == b"":  # break if end of file
                        self._Index__create_file(self._file,
                                                 self.header["SIZ"], num_atts)
                        raise StopIteration

                    output = (yield {attribute: raw})[attribute]

                    if output is None:
                        nwid = b"\xff" * 4
                        output = b""
                        nleng = 0
                    else:
                        nleng = len(output)
                        nwid = pack("I", nleng)

                    self._file[pos:pos + wid + 4] = b"".join([nwid, output])

                    pos += nleng + 4
                else:
                    wid = unpack("I", self._file.read(4))[0]
                    if wid == 4294967295:
                        wid = 0

                    self._file.seek(wid, 1)
                    pos += wid + 4
            x += 1

        self._Index__create_file(self._file, self.header["SIZ"], num_atts)

        raise StopIteration

    def select(self, attribute, start, stop):
        key = start
        atts = self.header["ATTBYID"]
        s = self._file.seek
        r = self._file.read

        pos = self._Index__get_position(key)  # Find pos

        if pos is False:
            raise IndexError("No key found for specified range")

        while key < stop:
            output = {}
            s(pos)

            # Read atts
            for att in atts:
                try:
                    wid = unpack("I", r(4))[0]
                    pos += 4
                except error:
                    raise StopIteration
                    break

                if wid != 4294967295:
                    pos += wid

                if atts[att][1] not in attribute:
                    if wid != 4294967295:
                        s(wid, 1)
                    continue

                if wid == 4294967295:
                    output[atts[att][1]] = None
                else:
                    output[atts[att][1]] = r(wid)

            key += 1
            yield output

        raise StopIteration

Register.register_fhandler(VARIABLE_WIDTH_FILE, location_is_file)
Register.default_fhandler = VARIABLE_WIDTH_FILE
