'''
This file is part of HTD (HayzenTech Database)

Copyright (C) 2011 - 2012 Andrew Hayzen

HTD is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

HTD is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with HTD. If not, see <http://www.gnu.org/licenses/>

Header object - Manages the I/O for the header
'''

from .. import DataType, Register
from .._objects.attribute import Attribute
from .._objects.user import User
from copy import deepcopy
from struct import pack, unpack
from time import time, sleep


class Header():
    def __init__(self):
        self.header = {"TMS": (0, 0)}

    def __load(self):
        """ Load header into cache """

        # Check data exists
        self._file.seek(0)
        data = self._file.read(8)

        if data == b"":
            return False

        # Read data
        length = unpack("Q", data)[0]
        file_format = self._file.read(1)
        timestamp, time_id = unpack("QB", self._file.read(9))
        header = {"FOR": file_format, "TMS": (timestamp, time_id)}
        pos = self._file.tell()

        while pos < length:
            wid = unpack("Q", self._file.read(8))[0]
            data = self._file.read(wid)
            header[str(data[0:3].decode("ISO-8859-1"))] = data[3:]
            pos += 8 + wid

        # Convert data

        # Convert Version
        header["VER"] = float(header["HTD"])
        del header["HTD"]

        # Convert each
        for att in header:
            if att not in ["VER", "FOR", "TMS"]:
                header[att] = eval(header[att])

            if att == "ATT":
                for a in deepcopy(header[att]):
                    val = header[att][a]

                    dtype = getattr(DataType,
                                    Register.dtype_codes[val["dtype"]])

                    size = val["size"]

                    if not isinstance(size, bytes):
                        size = size.encode("ISO-8859-1")

                    size = unpack("I", size)[0]

                    tmp = Attribute(dtype=dtype, size=size)
                    tmp._id = val["id"]
                    tmp._name = a
                    tmp._limit_list = val["limit_list"]
                    tmp._range = val["range"]
                    tmp._reg_expr = val["reg_expr"]
                    tmp._required = val["required"]
                    tmp._default = val["default"]

                    header[att][a] = tmp

            if att == "USR":
                for a in deepcopy(header[att]):
                    val = header[att][a]
                    header[att][a] = User(access=val["acc"])
                    header[att][a]._User__username = a
                    header[att][a]._User__password = val["pass"]  # sha256

        # Add calculated
        header["SIZ"] = length
        header["ATTBYID"] = self._get_atts_by_id(header["ATT"])

        # Update var
        self.header = header

        return self.header

    def __update(self, header, blank=False):  # Convert from new objects to raw
        """ Set new header """
        # Get old header
        if not blank:
            oldheader = deepcopy(self.header)

        # Build inital str
        leng = pack("Q", len(str(header["VER"])) + 3)

        write = [self.file_format]

        # Load timestamp
        while 1:
            timestamp = pack("Q", int(time() * 10 ** 6))
            if timestamp != self.header["TMS"][0]:
                time_id = b"\x01"
                break
            elif self.header["TMS"][1] == 255:
                sleep(0.1)
                continue
            else:
                time_id = pack("B", self.header["TMS"][1] + 1)
                break

        write.extend([timestamp, time_id, leng, b"HTD",
                      str(header["VER"]).encode("ISO-8859-1")])

        # Add header atts
        for att in ["KEY", "USR", "TAB", "ATT", "R1M", "RM1"]:
            if att in ["ATT", "USR"]:
                tmp = {}
                for i in header[att]:
                    tmp[i] = header[att][i].to_dict()
            else:
                tmp = header[att]

            length = len(str(tmp)) + 3
            write.extend([pack("Q", length), att.encode("ISO-8859-1"),
                          str(tmp).encode("ISO-8859-1")])

        write = b"".join(write)
        write = b"".join([pack("Q", len(write) + 8), write])  # Add length

        # Write to file
        if blank:
            self._file[0:1] = write
        else:
            self._file[0:oldheader["SIZ"]] = write

        self.__load()

        return True

    def _get_atts_by_id(self, atts):
        """ Sort attributes by id """
        attributes = {}

        for att in atts:
            attr = atts[att]
            attributes[attr._id] = [attr.size, att]

        return attributes

    def header_confirm(self):
        """ Check for changes in the header """
        self._file.seek(9)

        timestamp, time_id = unpack("QB", self._file.read(9))
        otimestamp, otime_id = self.header["TMS"]  # Load old times

        if otime_id != time_id or otimestamp != timestamp:
            self.header_update()

        return
